var Mutual = require('.')
var test = require('tape')
var pull = require('pull-stream')
var ssbKeys = require('ssb-keys')
var createFeed  = require('ssb-feed')
var createConfig  = require('ssb-config/inject')
var schemas = require('./schemas')

var createSbot = require('scuttlebot')
  .use(require('scuttlebot/plugins/private'))

var sbot = createSbot({
  temp: 'test-ssb-mutual', timeout: 1000,
  keys: ssbKeys.generate()
})

var alice = createFeed(sbot, ssbKeys.generate())
var bob = createFeed(sbot, ssbKeys.generate())
var carol = createFeed(sbot, ssbKeys.generate())
var community
var msg1, msg2, msg2, msg4, msg5, msg6, msg7

var mutual = Mutual.init(sbot)

test('zero balances', function (t) {
  t.plan(10)

  mutual.getAccountBalance({
    account: alice.id,
    currency: 'FOO'
  }, function (err, amount) {
    t.error(err, 'getAccountBalance')
    t.equals(amount, '0', 'zero balance')
  })

  mutual.getAccountBalances({account: alice.id}, function (err, balances) {
    t.error(err, 'getAccountBalances')
    t.deepEquals(balances, {}, 'no balances')
  })

  mutual.getCurrencyBalances({currency: 'FOO'}, function (err, balances) {
    t.error(err, 'getCurrencyBalances')
    t.deepEquals(balances, {}, 'no balances')
  })

  pull(
    mutual.streamAccountHistory({account: alice.id}),
    pull.collect(function (err, txs) {
      t.error(err, 'streamAccountHistory')
      t.deepEquals(txs, [], 'no transactions')
    })
  )

  pull(
    mutual.streamTransactions({account: alice.id}),
    pull.collect(function (err, txs) {
      t.error(err, 'streamTransactions')
      t.deepEquals(txs, [], 'no transactions')
    })
  )
})

test('transaction with self', function (t) {
  t.test('alice sends 1 ASD to self', function (t) {
    var value = schemas.credit(alice.id, 1, 'ASD', 'send to self')
    alice.add(value, function (err, msg) {
      msg1 = msg
      t.error(err, 'alice.add')
      t.comment(JSON.stringify(msg))
      t.end()
    })
  })

  t.test('alice still has 0 balance', function (t) {
    mutual.getAccountBalance({
      account: alice.id,
      currency: 'ASD'
    }, function (err, amount) {
      t.error(err, 'getAccountBalance')
      t.equals(amount, '0', 'zero balance')
      t.end()
    })
  })

  t.test('currency balances are zero', function (t) {
    mutual.getCurrencyBalances({
      currency: 'ASD'
    }, function (err, balances) {
      t.error(err, 'getCurrencyBalances')
      t.deepEquals(balances, {}, 'no balances')
      t.comment(JSON.stringify(balances))
      t.end()
    })
  })
})

test('send currency', function (t) {
  t.test('alice sends 1.5 FOO to bob', function (t) {
    var value = schemas.credit(bob.id, 1.5, 'FOO', 'send some foo')
    alice.add(value, function (err, msg) {
      msg2 = msg
      t.error(err, 'alice.add')
      t.comment(JSON.stringify(msg))
      t.end()
    })
  })

  t.test('alice has negative balance', function (t) {
    mutual.getAccountBalance({
      account: alice.id,
      currency: 'FOO'
    }, function (err, amount) {
      t.error(err, 'getAccountBalance')
      t.equals(amount, '-1.5', 'updated balance')
      t.end()
    })
  })

  t.test('bob has positive balance', function (t) {
    mutual.getAccountBalance({
      account: bob.id,
      currency: 'FOO'
    }, function (err, amount) {
      t.error(err, 'getAccountBalance')
      t.equals(amount, '1.5', 'updated balance')
      t.end()
    })
  })

  t.test('currency balances are updated', function (t) {
    var bals = {}
    bals[alice.id] = '-1.5'
    bals[bob.id] = '1.5'
    mutual.getCurrencyBalances({
      currency: 'FOO'
    }, function (err, balances) {
      t.error(err, 'getCurrencyBalances')
      t.deepEquals(balances, bals, 'balances')
      t.comment(JSON.stringify(balances))
      t.end()
    })
  })

  t.test('alice\'s account history includes transaction', function (t) {
    pull(
      mutual.streamAccountHistory({account: alice.id}),
      pull.collect(function (err, txs) {
        t.error(err, 'streamAccountHistory')
        t.deepEquals(txs, [
          {
            id: msg1.key,
            timestamp: msg1.value.timestamp,
            counterparty: alice.id,
            private: false,
            amount: '1',
            originated: true,
            currency: 'ASD',
            memo: 'send to self',
          },
          {
            id: msg2.key,
            timestamp: msg2.value.timestamp,
            counterparty: bob.id,
            private: false,
            amount: '-1.5',
            originated: true,
            currency: 'FOO',
            memo: 'send some foo',
          }
        ], 'transaction')
        t.end()
      })
    )
  })

  t.test('bob\'s account history includes transaction', function (t) {
    pull(
      mutual.streamAccountHistory({account: bob.id}),
      pull.collect(function (err, txs) {
        t.error(err, 'streamAccountHistory')
        t.deepEquals(txs, [{
          id: msg2.key,
          timestamp: msg2.value.timestamp,
          counterparty: alice.id,
          private: false,
          amount: '1.5',
          originated: false,
          currency: 'FOO',
          memo: 'send some foo',
        }], 'transaction')
        t.end()
      })
    )
  })

  t.test('transactions log includes transaction', function (t) {
    pull(
      mutual.streamTransactions(),
      pull.collect(function (err, txs) {
        t.error(err, 'streamTransactions')
        t.deepEquals(txs, [
          {
            id: msg1.key,
            timestamp: msg1.value.timestamp,
            author: alice.id,
            counterparty: alice.id,
            private: false,
            amount: '1',
            currency: 'ASD',
            memo: 'send to self',
          },
          {
            id: msg2.key,
            timestamp: msg2.value.timestamp,
            author: alice.id,
            counterparty: bob.id,
            private: false,
            amount: '1.5',
            currency: 'FOO',
            memo: 'send some foo',
          }
        ], 'transaction')
        t.end()
      })
    )
  })
})

test('send another currency', function (t) {
  t.test('bob credits 0.00000001 BLA to alice', function (t) {
    var value = schemas.credit(alice.id, '0.00000001', 'BLA')
    bob.add(value, function (err, msg) {
      msg3 = msg
      t.error(err, 'bob.add')
      t.comment(JSON.stringify(msg))
      t.end()
    })
  })

  t.test('currency balances are updated', function (t) {
    var bals = {}
    bals[bob.id] = '-0.00000001'
    bals[alice.id] = '0.00000001'
    mutual.getCurrencyBalances({
      currency: 'BLA'
    }, function (err, balances) {
      t.error(err, 'getCurrencyBalances')
      t.deepEquals(balances, bals, 'balances')
      t.comment(JSON.stringify(balances))
      t.end()
    })
  })

  t.test('bob\'s account history is updated', function (t) {
    pull(
      mutual.streamAccountHistory({account: bob.id}),
      pull.collect(function (err, txs) {
        t.error(err, 'streamAccountHistory')
        t.deepEquals(txs, [
          {
            id: msg2.key,
            timestamp: msg2.value.timestamp,
            counterparty: alice.id,
            private: false,
            amount: '1.5',
            originated: false,
            currency: 'FOO',
            memo: 'send some foo',
          },
          {
            id: msg3.key,
            timestamp: msg3.value.timestamp,
            counterparty: alice.id,
            private: false,
            amount: '-0.00000001',
            originated: true,
            currency: 'BLA',
            memo: undefined,
          }
        ], 'transaction')
        t.end()
      })
    )
  })
})

test('send in loop', function (t) {
  t.test('carol debits 1.5 FOO from bob', function (t) {
    var value = schemas.credit(bob.id, -1.5, 'FOO')
    carol.add(value, function (err, msg) {
      msg4 = msg
      t.error(err, 'carol.add')
      t.comment(JSON.stringify(msg))
      t.end()
    })
  })

  t.test('carol credits 2 FOO to alice', function (t) {
    var value = schemas.credit(alice.id, 2, 'FOO')
    carol.add(value, function (err, msg) {
      msg5 = msg
      t.error(err, 'carol.add')
      t.comment(JSON.stringify(msg))
      t.end()
    })
  })

  t.test('alice balance updated', function (t) {
    mutual.getAccountBalance({
      account: alice.id,
      currency: 'FOO'
    }, function (err, amount) {
      t.error(err, 'getAccountBalance')
      t.equals(amount, '0.5', 'balance')
      t.end()
    })
  })

  t.test('bob balance updated', function (t) {
    mutual.getAccountBalance({
      account: bob.id,
      currency: 'FOO'
    }, function (err, amount) {
      t.error(err, 'getAccountBalance')
      t.equals(amount, '0', 'balance')
      t.end()
    })
  })

  t.test('carol balance updated', function (t) {
    mutual.getAccountBalance({
      account: carol.id,
      currency: 'FOO'
    }, function (err, amount) {
      t.error(err, 'getAccountBalance')
      t.equals(amount, '-0.5', 'balance')
      t.end()
    })
  })

  t.test('currency balances are updated', function (t) {
    var bals = {}
    bals[alice.id] = '0.5'
    bals[carol.id] = '-0.5'
    mutual.getCurrencyBalances({
      currency: 'FOO'
    }, function (err, balances) {
      t.error(err, 'getCurrencyBalances')
      t.deepEquals(balances, bals, 'balances')
      t.comment(JSON.stringify(balances))
      t.end()
    })
  })

  t.test('transactions log includes transactions', function (t) {
    pull(
      mutual.streamTransactions(),
      pull.collect(function (err, txs) {
        t.error(err, 'streamTransactions')
        t.deepEquals(txs, [
          {
            id: msg1.key,
            timestamp: msg1.value.timestamp,
            author: alice.id,
            counterparty: alice.id,
            private: false,
            amount: '1',
            currency: 'ASD',
            memo: 'send to self',
          },
          {
            id: msg2.key,
            timestamp: msg2.value.timestamp,
            author: alice.id,
            counterparty: bob.id,
            private: false,
            amount: '1.5',
            currency: 'FOO',
            memo: 'send some foo',
          },
          {
            id: msg3.key,
            timestamp: msg3.value.timestamp,
            author: bob.id,
            counterparty: alice.id,
            private: false,
            amount: '0.00000001',
            currency: 'BLA',
            memo: undefined,
          },
          {
            id: msg4.key,
            timestamp: msg4.value.timestamp,
            author: carol.id,
            counterparty: bob.id,
            private: false,
            amount: '-1.5',
            currency: 'FOO',
            memo: undefined,
          },
          {
            id: msg5.key,
            timestamp: msg5.value.timestamp,
            author: carol.id,
            counterparty: alice.id,
            private: false,
            amount: '2',
            currency: 'FOO',
            memo: undefined,
          }
        ], 'transactions')
        t.end()
      })
    )
  })
})

test('send to virtual account', function (t) {
  t.test('create community virtual account', function (t) {
    var value = schemas.account('community', 'community funds')
    sbot.publish(value, function (err, msg) {
      community = msg
      t.error(err, 'publish')
      t.comment(JSON.stringify(msg))
      t.end()
    })
  })

  t.test('alice credits 0.5 FOO to community', function (t) {
    var value = schemas.credit(community.key, 0.5, 'FOO')
    alice.add(value, function (err, msg) {
      msg6 = msg
      t.error(err, 'alice.add')
      t.comment(JSON.stringify(msg))
      t.end()
    })
  })

  t.test('alice balance updated', function (t) {
    mutual.getAccountBalance({
      account: alice.id,
      currency: 'FOO'
    }, function (err, amount) {
      t.error(err, 'getAccountBalance')
      t.equals(amount, '0', 'balance')
      t.end()
    })
  })

  t.test('community balance updated', function (t) {
    mutual.getAccountBalance({
      account: community.key,
      currency: 'FOO'
    }, function (err, amount) {
      t.error(err, 'getAccountBalance')
      t.equals(amount, '0.5', 'balance')
      t.end()
    })
  })

  t.test('currency balances are updated', function (t) {
    var bals = {}
    bals[community.key] = '0.5'
    bals[carol.id] = '-0.5'
    mutual.getCurrencyBalances({
      currency: 'FOO'
    }, function (err, balances) {
      t.error(err, 'getCurrencyBalances')
      t.deepEquals(balances, bals, 'balances')
      t.comment(JSON.stringify(balances))
      t.end()
    })
  })
})

test('private transaction', function (t) {
  t.test('sbot credits 0.5 FOO to carol, privately', function (t) {
    var value = schemas.credit(carol.id, 0.5, 'FOO')
    sbot.private.publish(value, [sbot.id, carol.id], function (err, msg) {
      msg7 = msg
      t.error(err, 'sbot.private.publish')
      t.comment(JSON.stringify(msg))
      t.end()
    })
  })

  t.test('currency balances are updated', function (t) {
    var bals = {}
    bals[community.key] = '0.5'
    bals[sbot.id] = '-0.5'
    mutual.getCurrencyBalances({
      currency: 'FOO'
    }, function (err, balances) {
      t.error(err, 'getCurrencyBalances')
      t.deepEquals(balances, bals, 'balances')
      t.comment(JSON.stringify(balances))
      t.end()
    })
  })
})

test('close', function (t) {
  mutual.close(function (err) {
    t.error(err, 'close')
    t.end()
  })
})
